;;; hmac.el --- RFC2104 Hashed Message Authentication Codes

;;; Copyright (C) 1998 Simon Josefsson <jas@pdc.kth.se>

;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.

;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.

;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

;;; Commentary:

;;; This library is part of nnimap, http://vic20.dzp.se/gnus-imap/.

;;; This is a quick'n'dirty, low performance, implementation of RFC2104.
;;;
;;; Example:
;;;
;;; (require 'md5)
;;; (hmac 'md5 64 16 "Jefe" "what do ya want for nothing?")
;;; "750c783e6ab0b503eaa86e310a5db738"
;;;
;;; 64 is block length of hash function (64 for MD5 and SHA), 16 is
;;; resulting hash length (16 for MD5, 20 for SHA).
;;;
;;; Tested with Emacs 20.2 and XEmacs 20.3.

;;; Release history:
;;;
;;; 1998-08-16  initial release posted to gnu.emacs.sources
;;; 1998-08-17  use append instead of char-list-to-string
 
(require 'cl)

;; Magic character for inner HMAC round. 0x36 == 54 == '6'
(defconst hmac-ipad ?\x36)

;; Magic character for outer HMAC round. 0x5C == 92 == '\'
(defconst hmac-opad ?\x5C)

;; Not so magic character for padding the key. 0x00
(defconst hmac-zero ?\x00)

(if (not (fboundp 'hexl-hex-string-to-integer))
    (defun hexl-hex-string-to-integer (string)
      (string-to-number string 16)))

(defun hmac (hash block-length hash-length key text)
  (let* (;; if key is longer than B, reset it to HASH(key)
	 (key (if (> (length key) block-length) 
		  (funcall hash key) key))
	 (k_ipad (append key nil))
	 (k_opad (append key nil)))
    ;; zero pad k_ipad/k_opad
    (while (< (length k_ipad) block-length)
      (setq k_ipad (append k_ipad (list hmac-zero))))
    (while (< (length k_opad) block-length)
      (setq k_opad (append k_opad (list hmac-zero))))
    ;; XOR key with ipad/opad into k_ipad/k_opad
    (setq k_ipad (mapcar (lambda (c) (logxor c hmac-ipad)) k_ipad))
    (setq k_opad (mapcar (lambda (c) (logxor c hmac-opad)) k_opad))
    ;; perform inner hash
    (let ((first-round (funcall hash (concat k_ipad text)))
	  de-hexed)
      (while (< 0 (length first-round))
	(push (hexl-hex-string-to-integer (substring first-round -2)) de-hexed)
	(setq first-round (substring first-round 0 -2)))
      ;; perform outer hash
      (funcall hash (concat k_opad de-hexed)))))

(provide 'hmac)
